package xgmext

var Terms = []string{
	"from:",
	"to:",
	"cc:",
	"bcc:",
	"subject:",
	"label:",
	"deliveredto:",
	"category:primary",
	"category:social",
	"category:promotions",
	"category:updates",
	"category:forums",
	"category:reservations",
	"category:purchases",
	"has:",
	"has:attachment",
	"has:drive",
	"has:document",
	"has:spreadsheet",
	"has:presentation",
	"has:youtube",
	"list:",
	"filename:",
	"in:",
	"is:",
	"is:important",
	"is:read",
	"is:unread",
	"is:starred",
	"after:",
	"before:",
	"older:",
	"newer:",
	"older_than:",
	"newer_than:",
	"size:",
	"larger:",
	"smaller:",
	"rfc822msgid:",
	"OR",
	"AND",
	"AROUND",
}
